/*
 *      ISIS release V1.1, Dec. 1988
 *      Export restrictions apply
 */
/* task.h:  multitasking datastructures */

/* Task-related constants */
# define        words           *4              /* Bytes per word */
# define        kw              *4*1024         /* Bytes per k-word */
#ifndef SUN4
# define        DEF_TASKLEN         (4 kw)          /* size of task struct */
#else
# define        DEF_TASKLEN         (6 kw)          /* size of task struct */
#endif

/* Signal disciplines */
# define        SD              0               /* Signal delayed */
# define        SU              1               /* Signal urgent */

#include <setjmp.h>

#ifdef	CTHREADS
#define THREADS			1

#include <cthreads.h>

#define	MUTEX_T			mutex_t
#define MUTEX_ALLOC(m)		m = mutex_alloc()
#define CONDITION_T		condition_t
#define CONDITION_ALLOC(c)	c = condition_alloc()
#define CONDITION_FREE(c)	condition_free(c)
#define THREAD_T                cthread_t
#define MUTEX_LOCK(m)           mutex_lock(m);
#define MUTEX_UNLOCK(m)		mutex_unlock(m);
#define CONDITION_WAIT(c)       condition_wait(c,isis_mutex);
#define CONDITION_SIGNAL(c)     condition_signal(c);
#define THREAD_SELF(t)          t = cthread_self()
#define THREAD_FORK(f,a,s)      cthread_detach(cthread_fork(f,a))
#endif CTHREADS

#ifdef  SUNLWP 
#define THREADS                 1 
 
#include <lwp/lwp.h>
#include <lwp/stackdep.h>

#define MUTEX_T                 mon_t
#define MUTEX_ALLOC(m)          mon_create(&m)
#define CONDITION_T             cv_t
#define CONDITION_ALLOC(c)      cv_create(&c,isis_mutex)
#define CONDITION_FREE(c)       cv_destroy(c)
#define THREAD_T                caddr_t
#define MUTEX_LOCK(m)           mon_enter(m)
#define MUTEX_UNLOCK(m)         mon_exit(m)
#define CONDITION_WAIT(c)       cv_wait(c)
#define CONDITION_SIGNAL(c)     cv_notify(c)
#define THREAD_SELF(s)          { lwp_self(&lwp_tid); s = lwp_tid.thread_id; }
#define THREAD_FORK(f,a,s)      lwp_create(&lwp_tid,f,MINPRIO,0,s,1,a)

thread_t  lwp_tid;
#endif SUNLWP   

#ifdef  APOLLO 
#define THREADS                 1 
 
#include <apollo/base.h>
#include <apollo/task.h>
#include <apollo/mutex.h>

#define DEF_STACKLEN            64*1024

#define MUTEX_T                 mutex_$lock_rec_t
#define THREAD_T                task_$handle_t
#define MUTEX_ALLOC(m)          mutex_$init(&m)
#define MUTEX_LOCK(m)           mutex_$lock(&m, mutex_$wait_forever)
#define MUTEX_UNLOCK(m)         mutex_$unlock(&m)
#define CONDITION_T             mutex_$lock_rec_t
#define CONDITION_ALLOC(c)      { MUTEX_ALLOC(c); MUTEX_LOCK(c); }
#define CONDITION_FREE(c)       MUTEX_UNLOCK(c)
#define CONDITION_WAIT(c)       { MUTEX_UNLOCK(isis_mutex); MUTEX_LOCK(c); MUTEX_LOCK(isis_mutex); }
#define CONDITION_SIGNAL(c)     { MUTEX_UNLOCK(c); }
#define THREAD_SELF(s)          s = task_$get_handle()
#define THREAD_FORK(f,a,s)      task_$create((task_$routine_p)f,(void*)a,0,DEF_STACKLEN,1,task_$forever,0,&acr_status)

status_$t acr_status;

#endif APOLLO   

struct  task
{
        jmp_buf         task_env;                       /* Saved registers */
        int             env_pad[16];                    /* Some jmp_bufs are too small */
        event_id        task_eid;                       /* Current event id */
        int             task_msgid;                     /* RPC msgid number */
        int             task_ccmsgid;                   /* msgid number for cc_terminate */
        int             task_rval;                      /* To return from t_wait */
        int             (*task_routine)();              /* Routine called */
        int             task_sleep;                     /* How long? */
        int             task_entry;                     /* Entry number, if any */
        short           task_act;                       /* Activity id for this task, if any */
        short           task_flag;                      /* flags */
        char            *task_arg0;                     /* Arg passed */
        address         task_addr;                      /* Address of this task */
        message         *task_msg_arg;                  /* Message arg to routine, if any */
        condition       *task_queue;                    /* Waiting on this qnode */
        bitvec          task_watching;                  /* Sites this task is watching */
        condition       task_cond;                      /* For waiting */
        char            *task_waitingfor;               /* Why waiting */
        condition       task_mwant;                     /* Wants message */
        message         *task_msg;                      /* Special for indirect (iterated) addressing */
        qnode           *task_msgs;                     /* Message we got */
        qnode           *task_active;                   /* Node on active task qnode */
        address         *task_cohorts;                  /* Cohort list, if in cc alg. */
        address         task_truesender;                /* Also for cc alg */
#ifdef THREADS
        CONDITION_T     task_runme;                     /* Signal to run me... */
        THREAD_T        task_threadid;                  /* Thread's unique id */
	char            *task_name;                     /* Copy of cthread->name */
#endif THREADS
        char            *task_end;                      /* stack area, normal case */
        char            *task_data;                     /* stack area, normal case */
};

/* Field names for qnode nodes */
#define TA_ISTASK               1

/* Variables for managing the task table */
task                   *do_task_dequeue();
extern task            *isis_ctp;                      /* Current task */
condition              isis_runqueue;                  /* Runqueue */
condition              isis_tasks;                     /* Task qnode */
#define                my_eid                          isis_ctp->task_eid

/* Macros for fork routines */
# define t_fork(routine, arg)           isis_fork(routine, arg, NULLMP)
# define t_fork_urgent(routine, arg)    isis_fork_urgent(routine, arg, NULLMP)
# define t_fork_msg(routine, mp)        isis_fork(routine, mp, mp)
# define t_fork_urgent_msg(routine, mp) isis_fork_urgent(routine, mp, mp)

# define t_waiting(condp)               ((*(condp)) != 0)

#define TASK_START      0x01                            /* Set for the startup task */
#define TASK_LOGGED     0x02                            /* Set if logged */
#define TASK_INHIBIT    0x04                            /* Set if inhibits joins */
#define TASK_CONGESTED  0x08                            /* Congested and waiting */
#define TASK_XBYREF     0x10                            /* Current caller used pass-by-ref */
#define TASK_ABYREF     0x20                            /* Called used address passing by ref */
#define TASK_TEMP       0x40                            /* Just visiting */
#define TASK_ZOMBIE     0x100				/* A member of the walking dead */

#define    task_dequeue(cp)  ((*cp == 0)? &isis_scheduler: do_task_dequeue(cp))

#ifdef  THREADS
MUTEX_T	isis_mutex;					/* Held when in ISIS */

# define        ISIS_ENTER()	if(1) { THREAD_T self; THREAD_SELF(self); if(isis_ctp->task_threadid != self) thread_isis_enter(); } else
# define        ISIS_EXIT()	if(1) { if(isis_ctp->task_flag&TASK_TEMP) thread_isis_exit(); } else
# define        ISIS_RETURN(x)	if(1) { if(isis_ctp->task_flag&TASK_TEMP) thread_isis_exit(); return(x); } else

#else    THREADS

# define	ISIS_ENTER()
# define 	ISIS_EXIT()
# define 	ISIS_RETURN(x)    return(x)

#endif   THREADS
